#include <unistd.h>
#include <sys/types.h>

#include "FirewireSource.h"

// Static definition
class FirewireSource *FirewireSource::_instance = NULL;

FirewireSource::FirewireSource(  )
{
	if( !GetInstance() ) _instance = this;
	// make the buffer safe
	buffer = NULL;
	// and all the other flags, so we don't get a segfault at exit
	bRaw1394HandleCreated = false;
	bDc1394CameraCreated = false;

#ifdef USE_CARD_0
	nCard = 0;
	strcpy( device, "/dev/video1394-0" );
#endif
#ifdef USE_CARD_1
	nCard = 1;
	strcpy( device, "/dev/video1394/1" );
#endif
#ifdef USE_CARD_2
	nCard = 2;
	strcpy( device, "/dev/video1394/2" );
#endif
#ifdef USE_CARD_3
	nCard = 3;
	strcpy( device, "/dev/video1394/3" );
#endif

#ifdef CAMERA_1024x768_8BIT
	format = FORMAT_SVGA_NONCOMPRESSED_1;
	mode = MODE_1024x768_MONO;
	frameRate = FRAMERATE_15;
	bytesPerPixel = 1;
#endif

#ifdef CAMERA_1024x768_16BIT
	format = FORMAT_SVGA_NONCOMPRESSED_1;
	mode = MODE_1024x768_MONO16;
	frameRate = FRAMERATE_15;
	bytesPerPixel = 2;
#endif

#ifdef CAMERA_640x480_8BIT
	format = FORMAT_VGA_NONCOMPRESSED;
	mode = MODE_640x480_MONO;
	frameRate = FRAMERATE_15;
	bytesPerPixel = 1;
#endif

#ifdef CAMERA_640x480_16BIT
	format = FORMAT_VGA_NONCOMPRESSED;
	mode = MODE_640x480_MONO16;
	frameRate = FRAMERATE_30;
	bytesPerPixel = 2;
#endif

}

/**
 * \brief Check to see if there is a firewire card, or a firewire bus, and then a camera handle.
 *
 */
void FirewireSource::GetDeviceNodes()
{
	bRaw1394HandleCreated = false;
	bDc1394CameraCreated = false;
	isStereo = isStereoCapable = false;
	buffer = NULL;
	bufferSize = 0;

	/*-----------------------------------------------------------------------
     * Open ohci and assign handle
     * Note: the variable specifies which 1394 card to open.  If you have multiple cards
     * you may want to try with (1) or (2) etc.
     *----------------------------------------------------------------------- */
     std::cout << "1394 Card number to be used" << nCard << ".\n";
	raw1394Handle = dc1394_create_handle( nCard );
	if( raw1394Handle == NULL )
	{
		std::cerr << "Unable to aquire a raw1394 handle\n\n" "Please check \n"
				 	 "  - if the kernel modules `ieee1394',`raw1394' and `ohci1394' are loaded\n"
				 	 "  - if you have read/write access to /dev/raw1394\n";
		bRaw1394HandleCreated = false;
		this->isActive = false;
	}
	else
	{
		bRaw1394HandleCreated = true;

		/*-----------------------------------------------------------------------
	     *  get the camera nodes and describe them as we find them
	     *-----------------------------------------------------------------------*/
		numNodes = raw1394_get_nodecount( raw1394Handle );

	//     Note: set 3rd parameter to 0 if you do not want the camera details printed
		cameraNodes = dc1394_get_camera_nodes( raw1394Handle, &numCameras, 1 );
		std::cout << "CAMERAS FOUND: " << numCameras << std::endl;
		fflush( stdout );
		if( numCameras < 1 )
		{
			/*
			 * if you get this message and you have multiple 1394 cards you may
			 * want to try modifying the input parameter for dc1394_create_handle()
			 * above
			 */

			std::cerr << "No cameras found! ("<<numNodes<<" nodes on the bus)\n"
					 	 "  - could be you need to try a different 1394 device (modify code to fix)\n";
		}

		/*
		 * -----------------------------------------------------------------------
		 * to prevent the iso-transfer bug from raw1394 system, check if camera is
	 * highest node. For details see
	 * http://linux1394.sourceforge.net/faq.html#DCbusmgmt and
	 * http://sourceforge.net/tracker/index.php?func=detail&aid=435107&group_id=8157&atid=108157
	 * -----------------------------------------------------------------------
	 */
	if( cameraNodes[0] == numNodes - 1 )
	{
		std::cerr <<
				 "\n" "Sorry, your camera is the highest numbered node\n"
				 "of the bus, and has therefore become the root node.\n"
				 "The root node is responsible for maintaining \n"
				 "the timing of isochronous transactions on the IEEE \n"
				 "1394 bus.  However, if the root node is not cycle master \n"
				 "capable (it doesn't have to be), then isochronous \n"
				 "transactions will not work.  The host controller card is \n"
				 "cycle master capable, however, most cameras are not.\n" "\n"
				 "The quick solution is to add the parameter \n"
				 "attempt_root=1 when loading the OHCI driver as a \n"
				 "module.  So please do (as root):\n" "\n" "   rmmod ohci1394\n"
				 "   insmod ohci1394 attempt_root=1\n" "\n"
				 "for more information see the FAQ at \n"
				 "http://linux1394.sourceforge.net/faq.html#DCbusmgmt\n" "\n";
	}

	/**
	 *
	 * This block here is for stereo camera detection. enable it when we need it.
	 *
	 */
//	dc1394_camerainfo dc1394CameraInfo;
//	dc1394_get_camera_info( raw1394Handle, cameraNodes[0], &dc1394CameraInfo );
//	std::cout << "Model : " <<dc1394CameraInfo.model << "\n";
//	if( !strcmp( dc1394CameraInfo.model, "BumbleBee" ) )
//		isStereoCapable = true;

	/// Hack for downsampling
	nScale = 2;
	oldScale = 0;
	/// No one shot
	isOneShot = false;
	/**
	 * Active node is node 0 at the start, change this to choose one camera or the other on Ramius
	 */
	activeNode = 0;
	}
}

FirewireSource::~FirewireSource(  )
{
	if( buffer )
	{
		delete buffer;
		buffer = NULL;
	}
	if( bDc1394CameraCreated )
	{
		//std::cout<< "Destroying camera handle...\n";
		dc1394_dma_unlisten( raw1394Handle, &dc1394Camera );
		dc1394_dma_release_camera( raw1394Handle, &dc1394Camera );
	}

	if( bRaw1394HandleCreated )
	{
		//std::cout << "Destroying RAW1394 handle...\n";
		raw1394_destroy_handle( raw1394Handle );
	}
}

/**
 * \fn bool InitCamera
 *
 * \brief Camera initialization routine for FireWire devices
 *
 */

bool FirewireSource::InitCamera()
{
	this->GetDeviceNodes();
	if( !this->bRaw1394HandleCreated )
		return false;
	/// First, setup the capture mode
	camResult = this->SetupCaptureMode();
	/// Then get the Bayer pattern
	camResult &= this->GetBayerPattern();
	/// Finally start the iso transmission
	camResult &= this->StartStopISO();
	return camResult;
}

bool FirewireSource::IsCameraFree( char *device )
{
	FILE *devfile = popen( "/usr/sbin/lsof | grep 1394", "r" );

	if( !devfile )
	{
		bool oFlag = false;
		char str[81];

		while( fgets( str, 80, devfile ) )
		{
			oFlag = true;
		}
		return !oFlag;
	}
	else
		return true;
}

bool FirewireSource::SetupCaptureMode(  )
{
	/*-----------------------------------------------------------------------
     *  setup capture
     *----------------------------------------------------------------------- */
	unsigned int channel;
	unsigned int speed;

	if( dc1394_get_iso_channel_and_speed
		( raw1394Handle, cameraNodes[activeNode], &channel, &speed ) != DC1394_SUCCESS )
	{
		std::cerr << "Unable to get the iso channel number\n";
		return false;
	}

	// note: format, mode, frameRate and bytesPerPixel are all defined as
	// globals
	// in the header

	/*
	 * Setup the capture mode
	 */
	 
	int e = dc1394_dma_setup_capture( raw1394Handle,
									  cameraNodes[activeNode],
									  channel,
									  format,
									  mode,
									  SPEED_400,
									  frameRate,
									  8,	// number of buffers
									  1,	// drop frames
									  device,
									  &dc1394Camera );

	if( e != DC1394_SUCCESS )
	{
		std::cerr <<
				 "Unable to setup camera-\n"
				 "check code above line "<<__LINE__<<" of "<<__FILE__<<" to make sure\n"
				 "that the video mode,framerate and format are\n"
				 "supported by your camera\n";
		return false;
	}
	/// Get the frame dimensions, keeping in mind Bayer downsampling status
	image_width = dc1394Camera.frame_width / nScale;
	image_height = dc1394Camera.frame_height / nScale;
	this->image_depth = 3;
	bDc1394CameraCreated = true;
	return bDc1394CameraCreated;
}


bool FirewireSource::StartStopISO( bool start )
{
	if( start )
	{
		std::cout << "Starting iso transmission...\n";

		if( dc1394_start_iso_transmission( raw1394Handle, dc1394Camera.node ) !=
			DC1394_SUCCESS )
		{
			std::cerr << "Unable to start camera iso transmission\n";
			return false;
		}
	}
	else
	{
		std::cout << "Releasing DMA...\n";
		// dc1394_dma_done_with_buffer( &dc1394Camera );
	/*-----------------------------------------------------------------------
	 *  Stop data transmission
	 *----------------------------------------------------------------------- */
		std::cout << "Stopping iso transmission...\n";
		if( dc1394_stop_iso_transmission( raw1394Handle, dc1394Camera.node ) !=
			DC1394_SUCCESS )
		{
			std::cerr << "Couldn't stop the camera?\n";
			return false;
		}
	}
	return true;
}

/**
 * GetBayerPattern()
 *
 * Method to read the Bayer encoding pattern from the camera registers,
 * so that we do not have to read the values from a config file for each
 * particular type of camera.
 *
 * \returns True if successfully gets valid a Bayer tile pattern
 *
 * Junaed Sattar
 * February 22, 2006.
 */

bool FirewireSource::GetBayerPattern(  )
{
	std::cout << "Getting Bayer Pattern...";
	quadlet_t qValue;

	GetCameraControlRegister( raw1394Handle, cameraNodes[activeNode], 0x1040,	/* Bayer
																		 * Tile
																		 * Mapping
																		 * register
																		 */
							  &qValue );

	pattern = BAYER_PATTERN_BGGR;
	switch ( qValue )
	{
	  case 0x42474752:			// BGGR
		  pattern = BAYER_PATTERN_BGGR;
		  std::cout << "Choosing BAYER_PATTERN_BGGR. \n";
		  break;
	  case 0x47524247:			// GRBG
		  pattern = BAYER_PATTERN_GRBG;
		  std::cout << "Choosing BAYER_PATTERN_GRBG. \n";
		  break;
	  case 0x52474742:			// RGGB
		  pattern = BAYER_PATTERN_RGGB;
		  std::cout << "Choosing BAYER_PATTERN_RGGB. \n";
		  break;
	  case 0x47425247:			// GBRG
		  pattern = BAYER_PATTERN_GBRG;
		  std::cout << "Choosing BAYER_PATTERN_GBRG. \n";
		  break;
	  case 0x59595959:			// YYYY = BW
		  std::cerr << "Camera is black and white\n";
		  /// Save the depth here
		  this->image_depth = 1;
		  return false;
	  default:
		  std::cerr << "Camera BAYER_TILE_MAPPING register has an unexpected value.\n";//<< qValue << std::endl;
		  return false;
	}
	return true;
}

void FirewireSource::GetImageSize(int *pXsize, int *pYsize) {
	(*pXsize) = image_width;
	(*pYsize) = image_height;
}

bool FirewireSource::GetFrame()
{
	camResult = true;

	if( !isOneShot )
	{
		if( dc1394_dma_single_capture( &dc1394Camera ) != DC1394_SUCCESS )
		{
			std::cerr << "Unable to capture a frame\n";
			camResult = false;
			return camResult;
		}
		else
		{
			dc1394_dma_done_with_buffer( &dc1394Camera );
		}
	}
	else
	{
		/// First stop the camera
		this->StartStopISO( false );

		/// then flush the DMA ring buffer
		while ( dc1394_dma_single_capture_poll( &dc1394Camera ) == DC1394_SUCCESS )
		{
			dc1394_dma_done_with_buffer( &dc1394Camera );
		}
		/// Set the one-shot capture mode.
		dc1394_set_one_shot( raw1394Handle, cameraNodes[activeNode]);
		/// Restart the ISO transmission
		this->StartStopISO( true );
		/// Capture a single frame using DMA
		if( dc1394_dma_single_capture( &dc1394Camera ) != DC1394_SUCCESS )
		{
			std::cerr << "Unable to capture a frame\n";
			camResult = false;
			return camResult;
		}
		else
		{
			dc1394_dma_done_with_buffer( &dc1394Camera );
		}
		isOneShot = false;
	}
//	if( refCount > 1 )
//	{
//		/// Get the frame dimensions, keeping in mind Bayer downsampling status
//		image_width = dc1394Camera.frame_width / nScale;
//		image_height = dc1394Camera.frame_height / nScale;
//		this->image_depth = 3;
//	}

	// allocate enough space
	bufferSize = image_width * image_height;
	if( DirtyBuffer(nScale) )
	{
		if( buffer ) delete buffer;
		buffer = new unsigned char[3 * bufferSize];
	}

	if( !buffer )
	{
		std::cerr << "Could not allocate RGB buffer for color conversion (size = "<< (3*bufferSize) <<")\n";
		camResult = false;
	}

	unsigned char *src = ( unsigned char * ) dc1394Camera.capture_buffer;

	if( bytesPerPixel > 1 )
	{
		src = new unsigned char[bufferSize];

		if( !src )
		{
			std::cout << "Could not allocate copy buffer for color conversion (size = "<< bufferSize <<")\n";
			camResult = false;
		}
		unsigned char *captureBuffer =
			( unsigned char * ) dc1394Camera.capture_buffer;

		for( unsigned long i = 0; i < bufferSize; i++ )
		{
			src[i] = captureBuffer[i * bytesPerPixel];
		}
	}

	/*-----------------------------------------------------------------------
     *  convert to color image
     *----------------------------------------------------------------------- */
	if( nScale > 1 )
	{
		BayerDownsample( src, buffer, image_width, image_height, pattern );
	}
	else
	{
		std::cout << "BayerEdgeSense\n";
		/*BayerEdgeSense*/BayerSimple( src, buffer, image_width, image_height, pattern );
	}

//	if( isOneShot )
//	{
//		camResult = this->StartStopISO();
//	}
	/* Don't care about vxl
	image.set_to_memory( ( const vxl_byte * ) buffer, image_width, image_height, image_depth,
                                                                3, 3 * image_width, 1 );
	if( !image ) camResult = false; */

	return camResult;
}
/**
 * This method switches between the two front firewire cameras in the AQUA Robot.
 * Only applicable to AQUA 1.5. Camera's must be similar in property, since no checks
 * like that are made.
 *
 */

bool FirewireSource::SwitchNodes()
{
	if( numCameras != 3 )
		return false;

	activeNode ^= 1;
	std::cout << "Active node is now: "<< activeNode << std::endl ;
	this->StartStopISO( false );
	if ( this->InitCamera() ) return true;
	return false;
}

bool FirewireSource::ToggleStereoMode()
{
	if( !isStereoCapable ) return false;
	isStereo = !isStereo;
	if( isStereo )
	{
		mode = MODE_1024x768_MONO16;
		bytesPerPixel = 2;
		return true;
	}
	mode = MODE_1024x768_MONO;
	bytesPerPixel = 1;
	return false;
}



/**
 * \brief Method to switch the camera(s) to Format7 mode, i.e. the region-of-interest mode.
 *
 */
bool FirewireSource::ROIMode( int tx, int ty, int width, int height )
{
	if( !bRaw1394HandleCreated ) return false;
	if( !numCameras ) return false;

	unsigned int channel;
	unsigned int speed;

	if( dc1394_get_iso_channel_and_speed
			( raw1394Handle, cameraNodes[activeNode], &channel, &speed ) != DC1394_SUCCESS )
	{
		std::cerr << "Unable to get the iso channel number\n";
		return false;
	}

	// note: format, mode, frameRate and bytesPerPixel are all defined as
	// globals
	// in the header

	/*
	 * Setup the capture mode
	 */
	int e = dc1394_dma_setup_format7_capture( raw1394Handle,
			cameraNodes[activeNode],
			channel,
			MODE_FORMAT7_1,
			SPEED_400,
			QUERY_FROM_CAMERA,
			tx, ty, width, height,
			8,	// number of buffers
			1,	// drop frames
			device,
			&dc1394Camera );

	if( e != DC1394_SUCCESS )
	{
		std::cerr <<
			"Unable to setup camera in FORMAT7 mode-\n"
			"check code above line "<<__LINE__<<" of "<<__FILE__<<" to make sure\n"
			"that the video mode,framerate and format are\n"
			"supported by your camera\n";
		return false;
	}
	/// Get the frame dimensions, keeping in mind Bayer downsampling status
	image_width = dc1394Camera.frame_width / nScale;
	image_height = dc1394Camera.frame_height / nScale;
	this->image_depth = 3;
	bDc1394CameraCreated = true;
	return bDc1394CameraCreated;
}

